import logging
import sys

from . import cli
from . import io


logger = logging.getLogger("hop")


def _add_parser_args(parser):
    cli.add_client_opts(parser)
    cli.add_logging_opts(parser)

    parser.add_argument(
        "message", metavar="MESSAGE", nargs="*",
        help="File of messages to publish. (standard input is used if omitted)"
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=io.Deserializer.__members__,
        default=io.Deserializer.BLOB.name,
        help="Specify the message format. Defaults to BLOB for an unstructured message.",
    )
    parser.add_argument(
        "-t", "--test",
        action="store_true",
        help="Mark messages as test messages by adding a header with key '_test'."
    )


def _main(args):
    """Publish messages.

    """
    cli.set_up_logger(args)

    loader = io.Deserializer[args.format]
    stream = io.Stream(auth=(not args.no_auth))

    with stream.open(args.url, "w") as s:
        logger.info("publishing messages to stream")
        for message_file in args.message:
            message = loader.load_file(message_file)
            s.write(message)

        if len(args.message) == 0:
            messages = sys.stdin.read().splitlines()

            if messages:
                assert args.format == io.Deserializer.BLOB.name \
                    or args.format == io.Deserializer.JSON.name, \
                    "piping/redirection only allowed for BLOB and JSON formats"

                for message in messages:
                    s.write(loader.load(message), test=args.test)



import certifi
from collections.abc import Mapping
import csv
import errno
import getpass
import logging
import os
import re
import stat
import toml

from adc import auth

from . import configure
from . import cli


SASLMethod = auth.SASLMethod


logger = logging.getLogger("hop")


# thin wrapper over adc's SASLAuth to define
# different defaults used within Hopskotch
class Auth(auth.SASLAuth):
    """Attach SASL-based authentication to a client.

    Returns client-based auth options when called.

    Parameters
    ----------
    user : `str`
        Username to authenticate with.
    password : `str`
        Password to authenticate with.
    host : `str`, optional
        The name of the host for which this authentication is valid.
    ssl : `bool`, optional
        Whether to enable SSL (enabled by default).
    method : `SASLMethod`, optional
        The SASL method to authenticate. The default is SASLMethod.OAUTHBEARER
        if token_endpoint is provided, or SASLMethod.SCRAM_SHA_512 otherwise.
        See valid SASL methods in SASLMethod.
    ssl_ca_location : `str`, optional
        If using SSL via a self-signed cert, a path/location
        to the certificate.
    token_endpoint : `str`, optional
        The OpenID Connect token endpoint URL.
        Required for OAUTHBEARER / OpenID Connect, otherwise ignored.
    """

    def __init__(self, user, password, host="", ssl=True, method=None,
                 token_endpoint=None, **kwargs):
        if method is None and token_endpoint is None:
            method = SASLMethod.SCRAM_SHA_512
        super().__init__(user, password, ssl=ssl, method=method,
                         token_endpoint=token_endpoint, **kwargs)
        self._username = user
        self._hostname = host

    @property
    def username(self):
        """The username for this credential"""
        return self._username

    @property
    def password(self):
        """The password for this credential"""
        return (self._config.get("sasl.password")
                or self._config.get("sasl.oauthbearer.client.secret"))

    @property
    def hostname(self):
        """The hostname with which this creential is associated,
            or the empty string if the credential did not contain this information
        """
        return self._hostname

    @property
    def mechanism(self):
        """The authentication mechanism to use"""
        return self._config["sasl.mechanism"]

    @property
    def protocol(self):
        """The communication protocol to use"""
        return self._config["security.protocol"]

    @property
    def ssl(self):
        """Whether communication should be secured with SSL"""
        return self.protocol != "SASL_PLAINTEXT"

    @property
    def ssl_ca_location(self):
        """The location of the Certfificate Authority data used for SSL,
            or None if SSL is not enabled
        """
        if "ssl.ca.location" not in self._config:
            return None
        return self._config["ssl.ca.location"]

    @property
    def token_endpoint(self):
        """The OpenID Connect token endpoint,
            or None if OpenID Connect is not enabled
        """
        return self._config.get("sasl.oauthbearer.token.endpoint.url")

    def __eq__(self, other):
        return (self._username == other._username
                and self.password == other.password
                and self.hostname == other.hostname
                and self.mechanism == other.mechanism
                and self.protocol == other.protocol
                and self.ssl_ca_location == other.ssl_ca_location
                and self.token_endpoint == other.token_endpoint)


class AmbiguousCredentialError(RuntimeError):
    """An exception used to indicate that an attempt to select a credential from a collection
    did not result in a unique best match.

    Parameters
    ----------
    message : `str`
        A description of the ambiguity which was encountered.
    """

    def __init__(self, message):
        self.message = message
        super(RuntimeError, self).__init__(message)


def load_auth(config_file=None):
    """Configures Auth instances from a configuration file.

    Args:
        config_file: Path to a configuration file, loading from
            the default location if not given.

    Returns:
        A list of configured Auth instances.

    Raises:
        RuntimeError: The config file exists, but has unsafe permissions
                      and will not be read until they are corrected.
        KeyError: An error occurred parsing the configuration file.
        FileNotFoundError: The configuration file, either as specified
                           explicitly or found automatically, does not exist

    """
    if config_file is None:
        config_file = configure.get_config_path("auth")
        # If finding data automatically, as a fallback, look to see if the auth data is in the main
        # config in the old style.
        if not os.path.exists(config_file):
            main_config_file = configure.get_config_path("general")
            if os.path.exists(main_config_file):
                try:
                    return load_auth(main_config_file)
                except RuntimeError:
                    # if the main config file does not contain auth data, rewrite the error into a
                    # complaint that the auth config file does not exist, since that is what should
                    # be fixed
                    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), config_file)
            # if config.toml does not exist, continue and complain about lack of auth.toml,
            # which is the real issue

    if not os.path.exists(config_file):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), config_file)
    # check that the config file has suitable permissions
    config_props = os.stat(config_file)
    dangerous_perm_mask = 0o7077
    if (config_props.st_mode & dangerous_perm_mask) != 0:
        raise RuntimeError(f"{config_file} has unsafe permissions: {config_props.st_mode:o}\n"
                           "Please correct it to 0600")

    # load config
    with open(config_file, "r") as f:
        try:
            auth_data = toml.loads(f.read())["auth"]
        except KeyError:
            raise RuntimeError("configuration file has no auth section")
        except Exception as ex:
            raise RuntimeError(f"configuration file is not configured correctly: {ex}")

    return _interpret_auth_data(auth_data)


def prune_outdated_auth(config_file=None):
    """Remove auth data from a general configuration file.

    This can be needed when updating auth data which was read from the general config for backwards
    compatibility, but is then written out to the correct new location in a separate auth config,
    as is now proper. With no further action, this would leave a vestigial copy from before the
    update in the general config file, which would not be rewritten, so this function exists to
    perform the necessary rewrite.

    Args:
        config_file: Path to a configuration file, rewriting
            the default location if not given.

    Raises:
        RuntimeError: The config file is malformed.

    """
    if config_file is None:
        config_file = configure.get_config_path("general")
    if not os.path.exists(config_file):
        return  # nothing to do!
    with open(config_file, "r") as f:
        try:
            config_data = toml.loads(f.read())
        except Exception as ex:
            raise RuntimeError(f"configuration file {config_file} is malformed: {ex}")
    if "auth" in config_data:
        del config_data["auth"]
        # only overwrite if we made a change
        with open(config_file, "w") as f:
            toml.dump(config_data, f)


def _interpret_auth_data(auth_data):
    """Configures Auth instances given data read from a configuration file.

    Args:
        auth_data: The data obtained from the top-level 'auth' TOML key

    Returns:
        A list of configured Auth instances.

    Raises:
        KeyError: A required key is missing from a credential entry.

    """
    if isinstance(auth_data, Mapping):
        # upgrade old-style single dict configs to new-style list-of-dicts (with one item)
        auth_data = [auth_data]

    auth = []
    for config in auth_data:
        # translate config options
        host = ""
        ssl = True
        extra_kwargs = {}
        try:
            # SSL options
            if "protocol" in config and config["protocol"] == "SASL_PLAINTEXT":
                ssl = False
            elif "ssl_ca_location" in config:
                extra_kwargs["ssl_ca_location"] = config["ssl_ca_location"]

            # SASL options
            user = config["username"]
            password = config["password"]

            if "hostname" in config:
                host = config["hostname"]

            token_endpoint = config.get("token_endpoint")

            if "mechanism" in config:
                mechanism = config["mechanism"].replace("-", "_")
            elif token_endpoint:
                mechanism = "OAUTHBEARER"
            else:
                mechanism = "SCRAM_SHA_512"

        except KeyError as ke:
            raise RuntimeError("configuration file is not configured correctly: "
                               f"missing auth property {ke}")
        else:
            auth.append(Auth(user, password, host=host, ssl=ssl, method=SASLMethod[mechanism],
                             token_endpoint=token_endpoint, **extra_kwargs))
    return auth


def _decompose_host_port(hp):
    """
    Split the host subcomponent of a URI from any port subcomponent which is present

    Args:
        hp:  A string which contains a host and may also contain a port specification, in the syntax
             specified by RFC 3986.
    Returns:
        A tuple containing the host substring, and the port substring or None if no port was
        explicitly specified.
    """
    split_re = re.compile(r"^([^[:][^:]*|\[[^\]]+\])(:([0-9]+))?$")
    match = split_re.match(hp)
    if match is None:
        return ("", None)
    return (match.group(1), match.group(3))


def select_matching_auth(creds, hostname, username=None):
    """Selects the most appropriate credential to use when attempting to contact the given host.

    Args:
        creds: A list of configured Auth objects. These can be obtained from :meth:`load_auth`.
        hostname: The name of the host for which to select suitable credentials.
        username: `str`, optional
            The name of the credential to use.

    Returns:
        A single Auth object which should be used to authenticate.

    Raises:
        RuntimeError: No credentials matched.
        AmbiguousCredentialError: More than one credential matched.
    """

    matches = []
    inexact_matches = []
    exact_hostname_match = False
    inexact_hostname_match = False
    target_host, target_port = _decompose_host_port(hostname)
    default_port = "9092"

    for cred in creds:
        maybe_exact_hostname_match = False
        maybe_inexact_hostname_match = False
        # If the credential has a hostname, we require it to match.
        # If it doesn't, we have to assume it might suit the user's purposes.
        if len(cred.hostname) > 0:
            cred_host, cred_port = _decompose_host_port(cred.hostname)
            # If both or neither have a port specified, they can be directly compared
            if (target_port is None) == (cred_port is None):
                if cred.hostname != hostname:
                    continue
                else:
                    maybe_exact_hostname_match = True
            # If the hosts differ there is no match
            elif cred_host != target_host:
                continue
            # Otherwise, the host is the same, and only one has a port specified.
            # If that port is the default port, it matches with the implicit default port of the
            # other.
            elif target_port == default_port or cred_port == default_port:
                # the hostname does match, but this credential might be rejected for other reasons
                # so we cannot directly update the global exact_hostname_match
                maybe_exact_hostname_match = True
        else:
            maybe_inexact_hostname_match = True

        # If the user specified a specific username, we require that to be matched.
        if username is not None and cred.username != username:
            continue

        if maybe_exact_hostname_match:
            matches.append(cred)
        elif maybe_inexact_hostname_match:
            inexact_matches.append(cred)
        exact_hostname_match |= maybe_exact_hostname_match
        inexact_hostname_match |= maybe_inexact_hostname_match

    # If there were no exact matches, take the inexact matches
    if inexact_hostname_match and not exact_hostname_match:
        matches = inexact_matches

    if len(matches) == 0:
        err = f"No matching credential found for hostname '{hostname}'"
        if username is not None:
            err += f" with username '{username}'"
        raise RuntimeError(err)

    if len(matches) > 1:
        err = f"Ambiguous credentials found for hostname '{hostname}'"
        err += f" with username '{username}'" if username is not None \
            else " with no username specified"
        err += "\nMatched credentials:"
        for match in matches:
            err += f"\n  {match.username}"
            err += " which has no associated hostname" if len(match.hostname) == 0 \
                else f" for {match.hostname}"
        raise AmbiguousCredentialError(err)

    # At this point we should have exactly one match
    return matches[0]


def _validate_hostname(input: str):
    """Ensure that a value entered for a hostname looks plausible.

    The user is supposed to enter a hostname, possibly with a port, not a URL.
    However, since users are likely to enter something valid prefixed with a `kafka://` scheme,
    accept that with a warning.
    We do not use standard URL parsing since the input is not supposed to be a URL, and attempting
    to parse it as such when no scheme is present mostly just leads to ambiguous results.

    Args:
        input: A string entered by the user when prompted for a hostname.

    Returns:
        The sanitized hostname or the empty string if the user enetered nothing.

    Raises:
        RuntimeError: If the user input is not acceptable.
    """
    if len(input) == 0:
        return input
    name_re = re.compile(r"^(kafka://)?(([^[:/]+|\[[^\]/]+\])(:[0-9]*)?)$")
    match = name_re.match(input)
    if match is None:
        raise RuntimeError("Unable to parse hostname. "
                           "Please enter either `hostname` or `hostname:port`.")
    if match.group(1) is not None:
        logger.warning(f"Ignoring '{match.group(1)}' prefix on hostname")
    return match.group(2)


def read_new_credential(csv_file=None):
    """Import a credential from a CSV file or obtain it interactively from the user.

    Args:
        csv_file: Path to a file from which to read credential data in CSV format.
                  If unspecified, the user will be prompted to enter data instead.

    Returns:
        A configured `Auth` object containing the new credential.

    Raises:
        FileNotFoundError: If csv_file is not None and refers to a nonexistent path.
        KeyError: If csv_file is not None and the specified file does not contain either a username
                  or password field.
        RuntimeError: If csv_file is None and the interactively entered username or passwod is
                      empty.

    """
    options = {}
    if csv_file is None:
        logger.info("Generating configuration with user-specified username + password")
        username = input("Username: ")
        if len(username) == 0:
            raise RuntimeError("Username may not be empty")
        password = getpass.getpass()
        if len(password) == 0:
            raise RuntimeError("Password may not be empty")
        hostname = _validate_hostname(input("Hostname (may be empty): "))
        token_endpoint = input("Token endpoint (empty if not applicable): ") or None
    else:
        if os.path.exists(csv_file):
            with open(csv_file, "r") as f:
                reader = csv.DictReader(f)
                cred = next(reader)
                username = cred["username"]
                password = cred["password"]
                hostname = cred["hostname"] if "hostname" in cred else ""
                token_endpoint = cred.get("token_endpoint")
                if "mechanism" in cred:
                    options["method"] = cred["mechanism"].replace("-", "_")
                if "protocol" in cred:
                    options["ssl"] = cred["protocol"] != "SASL_PLAINTEXT"
                if "ssl_ca_location" in cred:
                    options["ssl_ca_location"] = cred["ssl_ca_location"]
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), csv_file)
    return Auth(username, password, hostname, token_endpoint=token_endpoint, **options)


def write_auth_data(config_file, credentials):
    """Write configuration file for the set of credentials.

    Creates containing directories as needed.

    Args:
        config_file: configuration file path
        credentials: list of `Auth` objects representing credentials to be stored

    """
    cred_list = []
    for cred in credentials:
        cred_dict = {"username": cred.username, "password": cred.password,
                     "protocol": cred.protocol, "mechanism": cred.mechanism,
                     "token_endpoint": cred.token_endpoint}
        if len(cred.hostname) > 0:
            cred_dict["hostname"] = cred.hostname
        # This is slightly subtle: certifi.where() is the default location to use for CA data if no
        # other is specified. It should always be available, because we specify certifi
        # (transitively) as a dependency. However, chances are significant that it may at any given
        # time be provided from within a virtual environment, which the user may later delete. The
        # auth data is stored properly in the user's XDG_CONFIG_HOME, so it can survive and be
        # shared across potentially many virtual environments. If any virtual environment is deleted
        # after a path specific to it has been captured in the auth file, that credential would
        # become unusable (without manually correcting the path). Since we can always use the
        # current certifi path at runtime, it is therefore better to just never record it in the
        # auth file.
        if cred.ssl_ca_location is not None and cred.ssl_ca_location != certifi.where():
            cred_dict["ssl_ca_location"] = cred.ssl_ca_location

        cred_list.append(cred_dict)

    os.makedirs(os.path.dirname(config_file), exist_ok=True)
    fd = os.open(config_file, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, stat.S_IRUSR | stat.S_IWUSR)
    with open(fd, "w") as f:
        toml.dump({"auth": cred_list}, f)
    logger.info(f"Wrote configuration to: {config_file}")


def list_credentials():
    """Display a list of all configured credentials.

    """
    creds = load_auth()
    max_username_len = max([len(c.username) for c in creds]) if len(creds) > 0 else 1
    long_format = f"{{:{max_username_len}}} for {{}}"
    for cred in creds:
        if len(cred.hostname) > 0:
            print(str.format(long_format, cred.username, cred.hostname))
        else:
            print(cred.username)
    if len(creds) == 0 and os.isatty(1):
        print("No credentials configured")


def add_credential(args):
    """Load a new credential and store it to the persistent configuration.

    Args:
        args: Command line options/arguments object.
              args.cred_file is taken as the path to a CSV file to import, or if None the user is
              prompted to enter a credential directly.
              args.force controls whether an existing credential with an identical name will be
              overwritten.
    """
    # first load any existing credentials
    try:
        creds = load_auth()
    except FileNotFoundError:
        # if no auth file exists we can just treat that as there being no credentials
        creds = []

    # next, load the new credential
    new_cred = read_new_credential(args.cred_file)

    # check for any conflicts between the new credential and an existing one
    conflicting_cred_idx = None
    for idx, cred in enumerate(creds):
        if cred.username == new_cred.username:
            if len(cred.hostname) > 0 and len(new_cred.hostname) > 0 \
                    and cred.hostname == new_cred.hostname:
                conflicting_cred_idx = idx
            elif len(cred.hostname) == 0 and len(new_cred.hostname) == 0:
                conflicting_cred_idx = idx
    if conflicting_cred_idx is not None:
        if args.force:
            creds[conflicting_cred_idx] = new_cred
        else:
            logger.error("Credential already exists; overwrite with --force")
            return
    else:
        creds.append(new_cred)
    write_auth_data(configure.get_config_path("auth"), creds)
    prune_outdated_auth()


def _construct_ambiguous_deletion_message(username, hostname, matches):
    """Create an error message for an ambiguous request to delete a credential.

    This function should only be used by `delete_credential`.

    Args:
        username: The username for the credential targeted for deletion.
        hostname: The hostname for the credential targeted for deletion, if any.
        matches: the collection of possibly matching credentials found by delete_credential.

    Returns:
        An error message string.

    """
    err = f"Ambiguous credentials found for username '{username}'"
    if hostname is not None:
        err += f" with hostname '{hostname}'\n"
    else:
        err += " with no hostname specified\n"
    err += "Matched credentials:"
    for match in matches:
        err += f"\n  {match.username}"
        if len(match.hostname) == 0:
            err += " which has no associated hostname"
        else:
            err += f" for {match.hostname}"
    return err


def delete_credential(name: str):
    """Delete a credential from the persistent configuration.

    Args:
        name: The username, or username and hostname separated by an '@' character of the credential
        to delete.

    Raises:
        RuntimeError: No credentials matched.
        AmbiguousCredentialError: More than one credential matched.
    """
    # first load any existing credentials
    try:
        creds = load_auth()
    except FileNotFoundError:
        # if no auth file exists we can just treat that as there being no credentials
        creds = []

    if '@' in name:
        username, hostname = name.split('@')
    else:
        username = name
        hostname = None

    # next, try to figure out which one we're supposed to remove
    matches = []
    match_indices = []

    for idx, cred in enumerate(creds):
        # the username must match
        if cred.username != username:
            continue
        # if specified, the hostname must match
        if hostname is not None and cred.hostname != hostname:
            continue

        matches.append(cred)
        match_indices.append(idx)

    if len(matches) == 0:
        err = f"No matching credential found with username '{username}'"
        if hostname is not None:
            err += f" with hostname '{hostname}'"
        raise RuntimeError(err)
    elif len(matches) > 1:
        raise AmbiguousCredentialError(
            _construct_ambiguous_deletion_message(username, hostname, matches))

    # At this point we should have exactly one match, which we can delete
    del creds[match_indices[0]]
    write_auth_data(configure.get_config_path("auth"), creds)
    prune_outdated_auth()


def _add_parser_args(parser):
    cli.add_logging_opts(parser)

    subparser = parser.add_subparsers(title="commands", metavar="<command>", dest="command")
    subparser.required = True

    subparser.add_parser("locate", help="display configuration path")

    subparser.add_parser("list", help="Display all stored credentials")

    add_cred_parser = subparser.add_parser("add", help="Load a credential, "
                                           "specified either via a CSV file or interactively")
    add_cred_parser.add_argument("cred_file", type=str, default=None, nargs='?',
                                 help="Import credentials from CSV file")
    add_cred_parser.add_argument("--force", action="store_true", default=False,
                                 help="If set, overwrite any existing credential when a new one "
                                 "with the same username and hostname is added")

    del_cred_parser = subparser.add_parser("remove", help="Delete a stored credential")
    del_cred_parser.add_argument("name", type=str,
                                 help="The name of the credential to delete. This will be treated "
                                 "as a username, unless it contains a @ character, in which case it"
                                 " will be split into a username and hostname.")


def _main(args):
    """Authentication configuration utilities.

    """
    cli.set_up_logger(args)

    if args.command == "locate":
        print(configure.get_config_path("auth"))
    elif args.command == "list":
        list_credentials()
    elif args.command == "add":
        add_credential(args)
    elif args.command == "remove":
        delete_credential(args.name)



from . import cli
from .io import list_topics


def _main(args):
    """List available topics.

    """
    valid_topics = list_topics(args.url, auth=not args.no_auth)
    if len(valid_topics) == 0:
        print("No accessible topics")
    else:
        print("Accessible topics:")
        for topic in sorted(valid_topics.keys()):
            print(f" {topic}")


def _add_parser_args(parser):
    cli.add_client_opts(parser)



import argparse


def append_subparser(subparser, cmd, func, formatter_class=None):
    """Attach a command to a subparser, generating help from docstrings.

    Args:
        subparser: An argparse-based subparser to attach a command to.
        cmd: The command name.
        func: The function to call for said command.
        formatter: Whether to apply custom formatting for sub-subparser commands.

    """
    assert func.__doc__, "empty docstring: {}".format(func)
    help_ = func.__doc__.split("\n")[0].lower().strip(".")
    desc = func.__doc__.strip()

    if formatter_class:
        parser = subparser.add_parser(
            cmd,
            help=help_,
            description=desc,
            formatter_class=formatter_class
        )
    else:
        parser = subparser.add_parser(cmd, help=help_, description=desc)

    parser.set_defaults(func=func)

    return parser


class SubcommandHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Custom formatter for displaying subcommands.

    """

    def _format_action(self, action):
        # fixes spacing/indent introduced by metavar in subparser
        self._action_max_length = 20
        parts = super(argparse.RawDescriptionHelpFormatter, self)._format_action(action)

        # removes the metavar description between title and commands
        if action.nargs == argparse.PARSER:
            parts = "\n".join(parts.split("\n")[1:])

        return parts



import dataclasses
import logging
import os
import toml
from typing import Optional
import warnings

from . import cli


logger = logging.getLogger("hop")


def get_config_path(type: str = "general"):
    """Determines the default location for auth configuration.

    Args:
        type: The type of configuration data for which the path should be looked up.
              Recognized types are 'general' and 'auth'.

    Returns:
        The path to the requested configuration file.

    Raises:
        ValueError: Unrecognized config type requested.

    """

    file_for_type = {
        "general": "config.toml",
        "auth": "auth.toml",
    }

    if type not in file_for_type:
        raise ValueError(f"Unknown configuration type: '{type}'")

    auth_filepath = os.path.join("hop", file_for_type[type])
    if "XDG_CONFIG_HOME" in os.environ:
        return os.path.join(os.getenv("XDG_CONFIG_HOME"), auth_filepath)
    else:
        return os.path.join(os.getenv("HOME"), ".config", auth_filepath)


@dataclasses.dataclass
class Config:
    """A container class for general configuration parameters
    """
    fetch_external: bool = True
    automatic_offload: bool = True

    def asdict(self):
        return dataclasses.asdict(self)

    @classmethod
    def load(cls, input):
        if hasattr(input, "read"):
            input = input.read()
        try:
            return cls(**toml.loads(input)["config"])
        except KeyError:
            raise RuntimeError("configuration data has no config section") from None
        except Exception as ex:
            raise RuntimeError(f"configuration data is malformed: {ex}")

    def save(self, output):
        toml.dump({"config": self.asdict()}, output)


def _parse_bool(s: str):
    """Interpret a range of possible truthy string values as booleans
    """
    s = s.lower()
    if s == "true" or s == "yes" or s == "on" or s == "1":
        return True
    if s == "false" or s == "no" or s == "off" or s == "0":
        return False
    raise ValueError(f"Invalid boolean value '{s}'")


# The mapping of types to functions used for parsing strings provided by the user into configuration
# values. When no entry for a given type exists in this object, the type's normal constructor should
# be used.
_conversion_functions = {
    bool: _parse_bool
}


def load_config(config_path: Optional[str] = None):
    """
    Load the configuration settings for the library combining, in order of increasing precedence:
    1. The built-in defaults
    2. Defaults stored in the configuration file
    3. Values currently set in the environment as environment variables

    Args:
        config_path: The path to the configuration file to load. If unspecified, the default
                     location is used.
    Return: A Config object with all current settings
    """
    # First, load the specified file, or the default file if not specified
    # If the file does not exist, use built-in defaults
    if config_path is None:
        config_path = get_config_path("general")
    if os.path.exists(config_path):
        try:
            file = open(config_path, 'r')
        except Exception:
            warnings.warn(f"Unable to open {config_path} for reading; using default configuration")
            config = Config()
        else:
            try:
                with file:
                    config = Config.load(file)
            except Exception as ex:
                raise RuntimeError(f"Error loading {config_path}") from ex
    else:
        config = Config()

    # Second, check for any environment variables and apply them if found
    env_var_prefix = "HOP_"

    for field in dataclasses.fields(config):
        var_name = (env_var_prefix + field.name).upper()
        value = os.getenv(var_name)
        if value is None:
            continue
        try:
            if field.type in _conversion_functions:
                value = _conversion_functions[field.type](value)
            else:
                value = field.type(value)
        except Exception as ex:
            raise RuntimeError(f"Error parsing configuration variable {var_name}") from ex
        setattr(config, field.name, value)

    return config


def write_config_value(name: str, value):
    """
    Store a configuration value to the configuation file.

    Args:
        name: The name of the parameter to store. Must be a valid parameter name.
        value: The parameter value to store. Must be the correct type for the specified parameter.
    """
    config_path = get_config_path("general")

    # Load any existing configuration
    # We do not use load_config because we want to avoid baking in any temporary settings
    # provided in environment variables, or any built-in defaults not explicitly chosen by the
    # user.
    if os.path.exists(config_path):
        try:
            config = toml.load(config_path)["config"]
        except KeyError:
            raise RuntimeError("{config_path} has no config section") from None
        except Exception as ex:
            raise RuntimeError(f"{config_path} is malformed: {ex}")
    else:
        config = {}

    # Set/overwrite the parameter specified by the user
    config[name] = value

    # Write the resulting configuration back out
    os.makedirs(os.path.dirname(config_path), exist_ok=True)
    with open(config_path, 'w') as output:
        toml.dump({"config": config}, output)

    logger.info(f"Wrote {name} = {value} to {config_path}")


def _add_parser_args(parser):
    cli.add_logging_opts(parser)

    subparser = parser.add_subparsers(title="commands", metavar="<command>", dest="command")
    subparser.required = True

    locate_parser = subparser.add_parser("locate", help="display configuration path")
    locate_parser.add_argument("-t", "--type", dest="type", default="general", required=False,
                               choices=["general", "auth"],
                               help="The type of configuration file to locate. "
                               "Defaults to general configuration.")

    subparser.add_parser("show", help="display current configuration")

    set_parser = subparser.add_parser("set", help="set default value for a configuration parameter")
    set_parser.add_argument("name",
                            choices=[field.name for field in dataclasses.fields(Config)],
                            help="The configuration parameter to set")
    set_parser.add_argument("value",
                            help="The parameter value to set")


def _main(args):
    """Configuration utilities.

    """
    cli.set_up_logger(args)

    if args.command == "locate":
        print(get_config_path(args.type))
    elif args.command == "show":
        config = load_config()
        for field in dataclasses.fields(config):
            print(f"{field.name}: {getattr(config, field.name)}")
    elif args.command == "set":
        # Collect data supplied by the user
        name = args.name
        # the argument parser should enforce that name is a value configuration parameter
        for field in dataclasses.fields(Config):
            if field.name == name:
                break
        value = args.value
        # sanitize/normalize the provided value
        if field.type in _conversion_functions:
            value = _conversion_functions[field.type](value)
        else:
            value = field.type(value)

        write_config_value(name, value)



import dataclasses
from io import StringIO
import pytest
import toml
import os
from unittest.mock import MagicMock, patch
from conftest import temp_config, temp_environ

from hop import configure


def test_get_config_path(tmpdir):
    with temp_environ(HOME=str(tmpdir)):
        if "XDG_CONFIG_HOME" in os.environ:
            # this change will revert at the end of the with block
            del os.environ["XDG_CONFIG_HOME"]

        # with HOME set but not XDG_CONFIG_HOME the config location should resolve to inside
        # ${HOME}/.config
        expected_path = os.path.join(tmpdir, ".config", "hop", "config.toml")
        config_loc = configure.get_config_path()
        assert config_loc == expected_path

        with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
            # with XDG_CONFIG_HOME set, no .config path component should be assumed
            expected_path = os.path.join(tmpdir, "hop", "config.toml")
            config_loc = configure.get_config_path()
            assert config_loc == expected_path


def test_get_config_path_auth(tmpdir):
    with temp_environ(HOME=str(tmpdir)):
        if "XDG_CONFIG_HOME" in os.environ:
            # this change will revert at the end of the with block
            del os.environ["XDG_CONFIG_HOME"]

        # with HOME set but not XDG_CONFIG_HOME the config location should resolve to inside
        # ${HOME}/.config
        expected_path = os.path.join(tmpdir, ".config", "hop", "auth.toml")
        config_loc = configure.get_config_path("auth")
        assert config_loc == expected_path

        with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
            # with XDG_CONFIG_HOME set, no .config path component should be assumed
            expected_path = os.path.join(tmpdir, "hop", "auth.toml")
            config_loc = configure.get_config_path("auth")
            assert config_loc == expected_path


def test_get_config_path_invalid(tmpdir):
    with pytest.raises(ValueError):
        configure.get_config_path("hairstyle")


def test_Config_load():
    # can load from a string
    c1 = configure.Config.load(StringIO("[config]"))
    for field in dataclasses.fields(configure.Config):
        assert getattr(c1, field.name) == field.default, \
            "default values should be used when not specified"

    non_default = StringIO()
    non_default.write("[config]\n")
    for field in dataclasses.fields(configure.Config):
        if field.type is bool:
            non_default.write(f"{field.name} = {str(not field.default).lower()}\n")
    c2 = configure.Config.load(non_default.getvalue())
    for field in dataclasses.fields(configure.Config):
        if field.type is bool:
            assert getattr(c2, field.name) != field.default, \
                "non-default values should be read correctly"
        else:
            assert getattr(c2, field.name) == field.default, \
                "default values should be used when not specified"

    with pytest.raises(RuntimeError) as ex:
        configure.Config.load("valid = false\n")
    assert "configuration data has no config section" in str(ex)

    with pytest.raises(RuntimeError) as ex:
        configure.Config.load('{"this": "is not TOML", "json": "yes"}')
    assert "configuration data is malformed" in str(ex)


def test_Config_roundtrip():
    out_buf = StringIO()
    c1 = configure.Config()
    # set some non-default values
    for field in dataclasses.fields(configure.Config):
        if field.type is bool:
            setattr(c1, field.name, not field.default)
    c1.save(out_buf)
    in_buf = StringIO(out_buf.getvalue())
    c2 = configure.Config.load(in_buf)
    assert c1 == c2


def test_parse_bool():
    assert configure._parse_bool("TrUe")
    assert configure._parse_bool("Yes")
    assert configure._parse_bool("oN")
    assert configure._parse_bool("1")
    assert not configure._parse_bool("fAlsE")
    assert not configure._parse_bool("No")
    assert not configure._parse_bool("oFF")
    assert not configure._parse_bool("0")
    with pytest.raises(ValueError) as ex:
        configure._parse_bool("fish")
    assert "Invalid boolean value" in str(ex)
    assert "fish" in str(ex)


def test_load_config(general_config, tmpdir):
    with temp_config(tmpdir, general_config) as conf_dir, temp_environ(XDG_CONFIG_HOME=conf_dir):
        config = configure.load_config()
        assert not config.fetch_external
        assert not config.automatic_offload

    # if there is no file, we should construct a default object
    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        config = configure.load_config()
        assert config == configure.Config()

    # when specified, we should load from a non-default path
    with temp_config(tmpdir, general_config) as conf_dir, temp_environ(XDG_CONFIG_HOME=conf_dir):
        alt_path = f"{tmpdir}/other.toml"
        with open(alt_path, 'w') as f:
            toml.dump({"config": {"fetch_external": not configure.Config.fetch_external}}, f)
        config = configure.load_config(alt_path)
        os.remove(alt_path)
        assert not config.fetch_external
        assert config.automatic_offload

    # if the specified file exists but cannot be read,
    # we should construct a default object and emit a warning
    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)), \
            patch("hop.configure.os.path.exists", MagicMock(return_value=True)), \
            patch("builtins.open", MagicMock(side_effect=PermissionError())):
        with pytest.warns(Warning):
            config = configure.load_config()
        assert config == configure.Config()

    # malformed data should cause a meaningful error
    with temp_config(tmpdir, "BAD\tTOML") as conf_dir, temp_environ(XDG_CONFIG_HOME=conf_dir):
        with pytest.raises(RuntimeError) as ex:
            config = configure.load_config()
        assert f"Error loading {conf_dir}" in str(ex)


def test_load_config_env(tmpdir):
    with temp_environ(XDG_CONFIG_HOME=str(tmpdir), HOP_FETCH_EXTERNAL="no"):
        config = configure.load_config()
        assert not config.fetch_external
        assert config.automatic_offload

    with temp_environ(XDG_CONFIG_HOME=str(tmpdir), HOP_FETCH_EXTERNAL="elk"):
        with pytest.raises(RuntimeError) as ex:
            config = configure.load_config()
        assert "Error parsing configuration variable HOP_FETCH_EXTERNAL" in str(ex)
        assert ex.value.__cause__ is not None
        assert "Invalid boolean value" in str(ex.value.__cause__)

    @dataclasses.dataclass
    class FakeConfig:
        number: int = 2

    with patch("hop.configure.Config", FakeConfig):
        with temp_environ(XDG_CONFIG_HOME=str(tmpdir), HOP_NUMBER="8"):
            config = configure.load_config()
            assert config.number == 8

        with temp_environ(XDG_CONFIG_HOME=str(tmpdir), HOP_NUMBER="rhino"):
            with pytest.raises(RuntimeError) as ex:
                config = configure.load_config()
            assert "Error parsing configuration variable HOP_NUMBER" in str(ex)
            assert ex.value.__cause__ is not None
            assert "invalid literal for int() with base 10: 'rhino'" in str(ex.value.__cause__)


def test_write_config_value(general_config, tmpdir):
    os.makedirs(f"{tmpdir}/hop", exist_ok=True)

    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        value = not configure.Config.fetch_external
        configure.write_config_value("fetch_external", value)
        config_path = configure.get_config_path("general")
        assert os.path.exists(config_path)
        with open(config_path, 'r') as f:
            data = toml.load(f)
        os.remove(config_path)
        assert "config" in data
        assert "fetch_external" in data["config"]
        assert data["config"]["fetch_external"] == value

    with temp_environ(XDG_CONFIG_HOME=str(tmpdir), HOP_AUTOMATIC_OFFLOAD="0"):
        value = not configure.Config.fetch_external
        configure.write_config_value("fetch_external", value)
        config_path = configure.get_config_path("general")
        assert os.path.exists(config_path)
        with open(config_path, 'r') as f:
            data = toml.load(f)
        os.remove(config_path)
        assert "config" in data
        assert "fetch_external" in data["config"]
        assert "automatic_offload" not in data["config"], "environment variables must be ignored"

    # Valid TOML with the wrong data should yield an error
    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        config_path = configure.get_config_path("general")
        with open(config_path, 'w') as f:
            toml.dump({"not-config": 17}, f)
        with open(config_path, 'r') as f:
            data = f.read()
        print(data)
        with pytest.raises(RuntimeError) as ex:
            configure.write_config_value("fetch_external", True)
        assert "has no config section" in str(ex)

    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        config_path = configure.get_config_path("general")
        with open(config_path, 'w') as f:
            f.write('{"this": "is not TOML", "json": "yes"}')
        with pytest.raises(RuntimeError) as ex:
            configure.write_config_value("fetch_external", True)
        assert "is malformed" in str(ex)


def test_no_command_configure(script_runner):
    warning_message = (
        "hop configure: error: the following arguments are required: <command>"
    )
    ret = script_runner.run(["hop", "configure"])
    assert not ret.success
    assert warning_message in ret.stderr



from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field
import email
import fastavro
from io import BytesIO
import json
import re
from typing import Any, Dict, List, Union
import collections.abc

import xmltodict

from . import plugins
from . import avro_utils

JSONType = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]


def format_name(cls):
    if hasattr(cls, "format_name"):
        return cls.format_name
    return cls.__name__.lower()


@dataclass
class MessageModel(ABC):
    """An abstract message model.

    """

    def __bytes__(self):
        """Produce the canonical representation of this message.
        """
        return json.dumps(asdict(self)).encode("utf-8")

    def serialize(self):
        """Wrap the message with its format and content, for transmission of Kafka.

        Returns:
            A dictionary with "format" and "content" keys.
            The value stored under "format" is the format label.
            The value stored under "content" is the actual encoded data.

        """
        # by default, encode using JSON
        return {"format": format_name(type(self)),
                "content": bytes(self),
                }

    @classmethod
    def deserialize(cls, data):
        """Unwrap a message produced by serialize() (the "content" value).

        Returns:
            An instance of the model class.

        """
        # corresponding to the default serialize implementation, upack from JSON
        return cls(**json.loads(data.decode("utf-8")))

    @classmethod
    def load_file(cls, filename):
        """Create a new message model from a file.

        Args:
            filename: The path to a file.

        Returns:
            The message model.

        """
        with open(filename, "r") as f:
            return cls.load(f)

    @classmethod
    @abstractmethod
    def load(cls, input_):
        """Create a new message model from a file object or string.
        This base implementation has no functionality and should not be called.

        Args:
            input_: A file object or string.

        Returns:
            The message model.

        """
        raise NotImplementedError("MessageModel.load() should not be called")


@dataclass
class VOEvent(MessageModel):
    """Defines a VOEvent 2.0 structure.

    Implements the schema defined by:
        http://www.ivoa.net/Documents/VOEvent/20110711/

    """

    ivorn: str
    role: str = "observation"
    version: str = "2.0"

    Who: dict = field(default_factory=dict)
    What: dict = field(default_factory=dict)
    WhereWhen: dict = field(default_factory=dict)
    How: dict = field(default_factory=dict)
    Why: dict = field(default_factory=dict)
    Citations: dict = field(default_factory=dict)
    Description: dict = field(default_factory=dict)
    Reference: dict = field(default_factory=dict)

    def __str__(self):
        return json.dumps(asdict(self), indent=2)

    def __bytes__(self):
        # TODO: this isn't really suitable, as the output should be the original XML format
        # That requires some massaging of the data to restore information discarded by xmltodict.
        return str(self).encode("utf-8")

    def serialize(self):
        """Wrap the message with its format and content.

        Returns:
            A dictionary with "format" and "content" keys.
            The value stored under "format" is the format label.
            The value stored under "content" is the actual encoded data.

        """
        # by default, encode using JSON
        return {"format": format_name(type(self)),
                "content": json.dumps(asdict(self)).encode("utf-8"),
                }

    @classmethod
    def load(cls, xml_input):
        """Create a new VOEvent from an XML-formatted VOEvent.

        Args:
            xml_input: A file object, string, or generator.

        Returns:
            The VOEvent.

        """
        vo = xmltodict.parse(xml_input, attr_prefix="")

        # enter root and remove XML-specific namespaces
        return cls(**{k: v for k, v in vo["voe:VOEvent"].items() if ":" not in k})

    @classmethod
    def load_file(cls, filename):
        """Create a new VOEvent from an XML-formatted VOEvent file.

        Args:
            filename: Name of the VOEvent file.

        Returns:
            The VOEvent.

        """
        with open(filename, "rb") as f:
            return cls.load(f)


@dataclass
class GCNTextNotice(MessageModel):
    """A GCN Notice in the plain text format.

    The original message data is stored in the raw property,
    and the parsed message is available via the fields property,
    which is a dictionary.
    """

    raw: bytes
    fields: dict

    def __bytes__(self):
        return self.raw

    @classmethod
    def deserialize(cls, data):
        fields = {}
        line_pattern = re.compile("([^:]*): *(.*)")
        last_key = None
        for line in data.decode("utf-8").strip().splitlines():
            m = re.match(line_pattern, line)
            if m is None:
                # if the line does not match the "key: value" format,
                # we assume it is a continuation of a previous value
                if last_key is not None:
                    fields[last_key] += '\n' + line.lstrip()
            else:
                # if the key is repeated,
                # we treat the new value as a continuation of the previous one
                last_key = m.group(1).strip().lower()
                value = m.group(2).strip()
                if last_key in fields:
                    fields[last_key] += '\n' + value
                else:
                    fields[last_key] = value

        return cls(raw=data, fields=fields)

    @classmethod
    def load(cls, input):
        if hasattr(input, "read"):
            raw = input.read()
        else:
            raw = input
        return cls.deserialize(raw)

    @classmethod
    def load_file(cls, filename):
        with open(filename, "rb") as f:
            return cls.load(f)


@dataclass
class GCNCircular(MessageModel):
    """Defines a GCN Circular structure.

    The parsed GCN circular is formatted as a dictionary with
    the following schema:

        {'headers': {'title': ..., 'number': ..., ...}, 'body': ...}

    """

    header: dict
    body: str
    format_name = "circular"

    def __str__(self):
        headers = [(name.upper() + ":").ljust(9) + val for name, val in self.header.items()]
        return "\n".join(headers + ["", self.body])

    def __bytes__(self):
        return str(self).encode("utf-8")

    def serialize(self):
        """Wrap the message with its format and content.

        Returns:
            A dictionary with "format" and "content" keys.
            The value stored under "format" is the format label.
            The value stored under "content" is the actual encoded data.

        """
        # by default, encode using JSON
        return {"format": format_name(type(self)),
                "content": json.dumps(asdict(self)).encode("utf-8"),
                }

    @classmethod
    def load(cls, email_input):
        """Create a new GCNCircular from an RFC 822 formatted circular.

        Args:
            email_input: A file object or string.

        Returns:
            The GCNCircular.

        """
        if hasattr(email_input, "read"):
            message = email.message_from_file(email_input)
        else:
            message = email.message_from_string(email_input)

        # format gcn circular into header/body
        return cls(
            header={title.lower(): content for title, content in message.items()},
            body=message.get_payload(),
        )


@dataclass
class Blob(MessageModel):
    """Defines an opaque message blob.

    """

    content: bytes

    def __str__(self):
        return str(self.content)

    def __bytes__(self):
        return self.content

    @classmethod
    def deserialize(cls, data):
        return cls(content=data)

    @classmethod
    def load(cls, blob_input):
        """Create a blob message from input data.

        Args:
            blob_input: The unstructured message data (bytes) or file object.

        Returns:
            The Blob.

        """
        if hasattr(blob_input, "read"):
            raw = blob_input.read()
        else:
            raw = blob_input
        return cls(content=raw)

    @classmethod
    def load_file(cls, filename):
        """Create a new message model from a file.

        Args:
            filename: The path to a file.

        Returns:
            The message model.

        """
        # arbitrary data should not be subjected to line-ending conversion, etc.
        with open(filename, "rb") as f:
            return cls.load(f)


@dataclass
class JSONBlob(MessageModel):
    """Defines an unformatted message blob.

    """

    content: JSONType
    format_name = "json"

    def __str__(self):
        return str(self.content)

    def __bytes__(self):
        return json.dumps(self.content).encode("utf-8")

    @classmethod
    def deserialize(cls, data):
        return cls(content=json.loads(data.decode("utf-8")))

    @classmethod
    def load(cls, blob_input):
        """Create a blob message from input text.

        Args:
            blob_input: The unstructured message text or file object.

        Returns:
            The Blob.

        """
        if hasattr(blob_input, "read"):
            raw = blob_input.read()
        else:
            raw = blob_input
        if isinstance(raw, bytes):
            raw = raw.decode("utf-8")
        return cls(content=json.loads(raw))


@dataclass
class AvroBlob(MessageModel):
    """Defines an unformatted message blob.

    """

    # serializing as Avro supports essentially the same types as JSON
    content: Union[JSONType, List[JSONType]]
    schema: dict = None
    format_name = "avro"
    single_record: bool = True

    def __init__(self, content: Union[JSONType, List[JSONType]], schema: dict = None,
                 single_record: bool = True):
        if not single_record and not isinstance(content, collections.abc.Sequence):
            raise TypeError("AvroBlob requires content to be a sequence of records when "
                            "single_record = False")
        self.content = content
        self.schema = schema
        self.single_record = single_record

    def __str__(self):
        return str(self.content)

    def __bytes__(self):
        if self.single_record:
            records = [self.content]
        else:
            records = self.content

        if self.schema is None:  # make up an ad-hoc schema
            self.schema = avro_utils.SchemaGenerator().find_common_type(records)

        buffer = BytesIO()
        fastavro.writer(buffer,
                        self.schema,
                        records
                        )
        return buffer.getvalue()

    @classmethod
    def _read_avro(cls, stream, single_record=True):
        extracted = []
        reader = fastavro.reader(stream)
        for record in reader:
            extracted.append(record)
        if single_record:
            if len(extracted) > 1:
                raise TypeError("AvroBlob requires content to be a single "
                                "record when single_record = True, number of "
                                "records in message being deserialized is "
                                f"{len(extracted)}.")
            content = extracted[0]
        else:
            content = extracted
        return cls(content=content, schema=reader.writer_schema, single_record=single_record)

    @classmethod
    def deserialize(cls, data, single_record=True):
        return cls._read_avro(BytesIO(data), single_record=single_record)

    @classmethod
    def load(cls, blob_input, single_record=True):
        """Create a blob message from input avro data.

        Args:
            blob_input: The encoded Avro data or file object.
            single_record: True if input avro data only contains one record.
              True by default.

        Returns:
            The Blob.

        """
        if hasattr(blob_input, "read"):
            raw = blob_input
        else:
            if not isinstance(blob_input, bytes):
                raise TypeError
            raw = BytesIO(blob_input)
        return cls._read_avro(raw, single_record=single_record)

    @classmethod
    def load_file(cls, filename, single_record=True):
        """Create a new message model from a file.

        Args:
            filename: The path to a file.
            single_record: True if input avro data only contains one record.
              True by default.

        Returns:
            The message model.

        """
        with open(filename, "rb") as f:
            return cls.load(f, single_record=single_record)

    def __eq__(self, other):
        if type(self) is not type(other):
            return False
        # compare only content, not schemas
        return self.content == other.content

    def __hash__(self):
        raise NotImplementedError("AvroBlob objects are not hashable")


@dataclass
class ExternalMessage(MessageModel):
    """Defines a message which refers to data stored externally at some URL
    """

    url: str
    format_name = "external"

    @classmethod
    def load(cls, input):
        """Create a blob message from input text.

        Args:
            blob_input: The unstructured message text or file object.

        Returns:
            The Blob.

        """
        if hasattr(input, "read"):
            raw = input.read()
        else:
            raw = input
        if isinstance(raw, bytes):
            raw = raw.decode("utf-8")
        decoded = json.loads(raw)
        return cls(url=decoded["url"])


@plugins.register
def get_models():
    model_classes = [
        VOEvent,
        GCNTextNotice,
        GCNCircular,
        Blob,
        JSONBlob,
        AvroBlob,
        ExternalMessage,
    ]
    return {format_name(cls): cls for cls in model_classes}



from unittest.mock import patch, mock_open, MagicMock
import sys
import pytest
import json
from io import BytesIO, StringIO
from uuid import uuid4

from hop import __version__, configure, io, models
from conftest import temp_auth, temp_config, temp_environ


@pytest.mark.script_launch_mode("subprocess")
def test_cli_hop(script_runner, auth_config, tmpdir):
    ret = script_runner.run(["hop", "--help"])
    assert ret.success

    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        ret = script_runner.run(["hop", "--version"])
        assert ret.success

        assert f"hop version {__version__}\n" in ret.stdout
        assert ret.stderr == ""


@pytest.mark.script_launch_mode("subprocess")
def test_cli_hop_module(script_runner, auth_config, tmpdir):
    ret = script_runner.run(["python", "-m", "hop", "--help"])
    assert ret.success

    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        ret = script_runner.run(["python", "-m", "hop", "--version"])
        assert ret.success

        assert f"hop version {__version__}\n" in ret.stdout
        assert ret.stderr == ""


@pytest.mark.parametrize("message_format", ["voevent", "circular", "blob"])
def test_cli_publish(script_runner, message_format, message_parameters_dict):
    if sys.version_info < (3, 7, 4):
        if message_format == "voevent":
            pytest.skip("requires python3.7.4 or higher")

    ret = script_runner.run(["hop", "publish", "--help"])
    assert ret.success

    # load parameters from conftest
    message_parameters = message_parameters_dict[message_format]

    test_file = message_parameters["test_file"]
    model_text = message_parameters["model_text"]

    # test publishing files
    message_mock = mock_open(read_data=model_text)
    with patch("hop.models.open", message_mock) as mock_file, patch(
        "hop.io.Stream.open", mock_open()
    ) as mock_stream:

        broker_url = "kafka://hostname:port/message"
        ret = script_runner.run([
            "hop", "publish", broker_url, test_file, "--quiet",
            "-f", message_format.upper(), "--no-auth",
        ])

        # verify CLI output
        assert ret.success
        assert ret.stderr == ""

        # verify message was processed
        if message_format in ["voevent", "blob"]:
            mock_file.assert_called_with(test_file, "rb")
        else:
            mock_file.assert_called_with(test_file, "r")

        mock_stream.assert_called_with(broker_url, "w")

    # test publishing from stdin
    with patch("hop.io.Stream.open", mock_open()) as mock_stream:
        ret = script_runner.run(["hop", "publish", "-f", message_format.upper(), broker_url],
                                stdin=StringIO('"message1"\n"message2"'))
        if message_format == "blob":
            assert ret.success
        else:  # only the blob format is supported, others should trigger an error
            assert not ret.success
            assert "piping/redirection only allowed for BLOB and JSON formats" in ret.stderr


def test_cli_publish_blob_msgs(mock_broker, mock_producer, mock_consumer, mock_admin_client):
    from hop import publish, io, models
    args = MagicMock()
    args.url = "kafka://hostname:port/topic"
    args.format = io.Deserializer.BLOB.name
    args.test = False
    start_at = io.StartPosition.EARLIEST
    read_url = "kafka://group@hostname:port/topic"
    fixed_uuid = uuid4()
    fake_uuid = MagicMock()
    fake_uuid.uuid4 = MagicMock(return_value=fixed_uuid)

    mock_adc_producer = mock_producer(mock_broker, "topic")
    mock_adc_consumer = mock_consumer(mock_broker, "topic", "group")
    msgs = [b"a string", b"\x10\x00\x20\x0B"]
    for msg in msgs:
        with patch("sys.stdin", BytesIO(msg)) as mock_stdin, \
                patch("hop.io.adc_producer.Producer", return_value=mock_adc_producer), \
                patch("hop.io.consumer.Consumer", return_value=mock_adc_consumer), \
                patch("hop.io.AdminClient", return_value=mock_admin_client(mock_broker)), \
                patch("hop.io.uuid", fake_uuid):
            publish._main(args)

            # each published message should be on the broker
            encoded = models.Blob(msg).serialize()
            expected_msg = {
                "message": encoded["content"],
                "headers": [("_id", fixed_uuid.bytes),
                            ("_format", encoded["format"].encode("utf-8"))],
            }
            assert mock_broker.has_message("topic", **expected_msg)

            # reading from the broker should yield messages which match the originals
            with io.Stream(start_at=None, auth=False).open(read_url, "r") as s:
                extracted_msgs = []
                for extracted_msg in s:
                    extracted_msgs.append(extracted_msg)
                # there should be one new message
                assert len(extracted_msgs) == 1
                # and it should be the one we published
                assert msg in [msg.content for msg in extracted_msgs]


def test_cli_publish_json_blob_msgs(mock_broker, mock_producer, mock_consumer, mock_admin_client):
    from hop import publish, io, models
    args = MagicMock()
    args.url = "kafka://hostname:port/topic"
    args.format = io.Deserializer.JSON.name
    args.test = False
    start_at = io.StartPosition.EARLIEST
    read_url = "kafka://group@hostname:port/topic"
    fixed_uuid = uuid4()
    fake_uuid = MagicMock()
    fake_uuid.uuid4 = MagicMock(return_value=fixed_uuid)

    mock_adc_producer = mock_producer(mock_broker, "topic")
    mock_adc_consumer = mock_consumer(mock_broker, "topic", "group")
    msgs = ["a string", ["a", "list", "of", "values"],
            {"a": "dict", "with": ["multiple", "values"]}]
    for msg in msgs:
        with patch("sys.stdin", StringIO(json.dumps(msg))) as mock_stdin, \
                patch("hop.io.adc_producer.Producer", return_value=mock_adc_producer), \
                patch("hop.io.consumer.Consumer", return_value=mock_adc_consumer), \
                patch("hop.io.AdminClient", return_value=mock_admin_client(mock_broker)), \
                patch("hop.io.uuid", fake_uuid):
            publish._main(args)

            # each published message should be on the broker
            encoded = models.JSONBlob(msg).serialize()
            expected_msg = {
                "message": encoded["content"],
                "headers": [("_id", fixed_uuid.bytes),
                            ("_format", encoded["format"].encode("utf-8"))],
            }
            assert mock_broker.has_message("topic", **expected_msg)

            # reading from the broker should yield messages which match the originals
            with io.Stream(start_at=None, auth=False).open(read_url, "r") as s:
                extracted_msgs = []
                for extracted_msg in s:
                    extracted_msgs.append(extracted_msg)
                # there should be one new message
                assert len(extracted_msgs) == 1
                # and it should be the one we published
                assert msg in [msg.content for msg in extracted_msgs]


def test_cli_publish_bad_blob(mock_broker, mock_producer, mock_admin_client):
    # ensure that invalid JSON causes an exception to be raised
    from hop import publish, io

    args = MagicMock()
    args.url = "kafka://hostname:port/topic"
    args.format = io.Deserializer.JSON.name
    args.test = False

    mock_adc_producer = mock_producer(mock_broker, "topic")
    msgs = ["not quoted", '{"unclosed:"brace"',
            "invalid\tcharacters\\\b"]
    for msg in msgs:
        # note that we do not serialize the messages as JSON
        with patch("sys.stdin", StringIO(msg)) as mock_stdin, \
                patch("hop.io.adc_producer.Producer", return_value=mock_adc_producer), \
                patch("hop.io.AdminClient", return_value=mock_admin_client(mock_broker)), \
                pytest.raises(ValueError):
            publish._main(args)


# create a mock output stream which supports writing binary data
def make_mock_stream():
    # this will be the underlying object which holds the data
    buffer = BytesIO()
    # Prepare wrapper routines which take strings like a TextIOBase.
    # Currently only acting as an output stream is supported.
    encoding = "utf-8"

    def write(s, /):
        encoded = s.encode(encoding)
        buffer.write(encoded)
        return len(encoded)
    # this will return bytes, not str, because the data is not constrained by any string encoding

    def getvalue():
        return buffer.getvalue()
    stream = MagicMock()
    stream.buffer = buffer
    stream.write = write
    stream.getvalue = getvalue
    return stream


def test_cli_subscribe(script_runner):
    with patch("pytest_console_scripts.StreamMock", make_mock_stream):
        ret = script_runner.run(["hop", "subscribe", "--help"])
        assert ret.success

        with patch("hop.io.Stream.open", mock_open()) as mock_stream:

            broker_url = "kafka://hostname:port/message"
            ret = script_runner.run(["hop", "subscribe", broker_url, "--no-auth"])

            # verify CLI output
            assert ret.success
            assert len(ret.stderr) == 0

            # verify broker url was processed
            mock_stream.assert_called_with(broker_url, "r", group_id=None, ignoretest=True)

            ret = script_runner.run(["hop", "subscribe", broker_url, "--no-auth", "--group-id",
                                     "group"])

            # verify CLI output
            assert ret.success
            assert len(ret.stderr) == 0

            # verify consumer group ID was used
            mock_stream.assert_called_with(broker_url, "r", group_id="group", ignoretest=True)

        message_body = "some-message"
        message_model = models.Blob(message_body.encode("utf-8"))
        message = io.Consumer.ExternalMessage(*io.Producer.pack(message_model), "a_topic",
                                              0, 0, 0, None)
        mock_instance = MagicMock()
        mock_instance.stream = MagicMock(return_value=[message])
        with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)):
            ret = script_runner.run(["hop", "--debug", "subscribe", broker_url, "--no-auth",
                                     "--quiet"])
            assert ret.success
            assert len(ret.stderr) == 0
            assert message.value() in ret.stdout

        message = io.Consumer.ExternalMessage(*io.Producer.pack(message_model, test=True),
                                              "a_topic", 0, 0, 0, None)
        mock_instance.stream = MagicMock(return_value=[message])

        with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)):
            ret = script_runner.run(["hop", "subscribe", broker_url, "--no-auth", "--quiet"])
            assert ret.success
            assert len(ret.stderr) == 0
            assert len(ret.stdout) == 0

        with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)):
            ret = script_runner.run(["hop", "subscribe", broker_url, "--no-auth", "--quiet",
                                     "--test"])
            assert ret.success
            assert len(ret.stderr) == 0
            assert message.value() in ret.stdout


def test_cli_subscribe_logging(script_runner):
    broker_url = "kafka://hostname:port/message"
    message_body = "some-message"
    message_model = models.Blob(message_body.encode("utf-8"))
    message = io.Consumer.ExternalMessage(*io.Producer.pack(message_model), "a_topic",
                                          0, 0, 0, None)
    mock_instance = MagicMock()
    mock_instance.stream = MagicMock(return_value=[message])
    with patch("hop.io.consumer.Consumer", MagicMock(return_value=mock_instance)), \
            patch("pytest_console_scripts.StreamMock", make_mock_stream):
        # check logging with --quiet (only warnings/errors)
        ret = script_runner.run(["hop", "--debug", "subscribe", broker_url, "--no-auth", "--quiet"])
        assert ret.success
        assert b"DEBUG" not in ret.stderr
        assert b"INFO" not in ret.stderr

        # check default logging (INFO)
        ret = script_runner.run(["hop", "--debug", "subscribe", broker_url, "--no-auth"])
        assert ret.success
        assert b"INFO" in ret.stderr
        assert b"DEBUG" not in ret.stderr

        # check verbose logging (DEBUG)
        ret = script_runner.run(["hop", "--debug", "subscribe", broker_url, "--no-auth", "-v"])
        assert ret.success
        assert b"INFO" in ret.stderr
        assert b"DEBUG" in ret.stderr


def test_cli_interrupt(script_runner):
    mock_consumer = MagicMock()
    interrupt = MagicMock(side_effect=KeyboardInterrupt())
    mock_consumer.__enter__ = interrupt
    mock_consumer.__iter__ = interrupt
    mock_consumer.__exit__ = interrupt
    with patch("hop.io.Stream.open", MagicMock(return_value=mock_consumer)):
        broker_url = "kafka://hostname:port/topic"
        ret = script_runner.run(["hop", "subscribe", broker_url, "--no-auth"])
        assert ret.success
        assert "INFO" in ret.stderr
        assert "received keyboard interrupt, closing" in ret.stderr


def dummy_topic_info(topic, error=None):
    info = MagicMock()
    info.error = error
    info.partitions = {}
    if error is None:
        info.partitions[0] = MagicMock()
    return info


def make_consumer_mock(expected_topics):
    metadata = MagicMock()
    metadata.topics = expected_topics
    list_topics = MagicMock(return_value=metadata)
    consumer = MagicMock()
    consumer.list_topics = list_topics
    return MagicMock(return_value=consumer)


def test_cli_list_topics(script_runner, auth_config, tmpdir):
    ret = script_runner.run(["hop", "list-topics", "--help"])
    assert ret.success

    broker_url = "kafka://hostname:port/"

    # general listing when no topics are returned
    with patch("confluent_kafka.Consumer", make_consumer_mock({})) as mock_consumer:
        ret = script_runner.run(["hop", "list-topics", broker_url, "--no-auth"])

        assert ret.success
        assert ret.stderr == ""
        assert "No accessible topics" in ret.stdout

        mock_consumer.assert_called()
        mock_consumer.return_value.list_topics.assert_called_with(timeout=-1.)

    expected_topics = ["foo", "bar"]
    unexpected_topics = ["baz"]
    topic_results = {}
    for topic in expected_topics:
        topic_results[topic] = dummy_topic_info(topic)
    for topic in unexpected_topics:
        topic_results[topic] = dummy_topic_info(topic, "an error")

    # general listing when some topics are returned
    with patch("confluent_kafka.Consumer", make_consumer_mock(topic_results)) as mock_consumer:
        ret = script_runner.run(["hop", "--debug", "list-topics", broker_url, "--no-auth"])

        assert ret.success
        assert ret.stderr == ""
        assert "Accessible topics" in ret.stdout
        for topic in expected_topics:
            assert topic in ret.stdout
        for topic in unexpected_topics:
            assert topic not in ret.stdout

        mock_consumer.assert_called()
        mock_consumer.return_value.list_topics.assert_called_with(timeout=-1.)

    query_topics = ["foo", "bar", "baz"]
    # listing of specific topics, none of which exist
    with patch("confluent_kafka.Consumer", make_consumer_mock({})) as mock_consumer:
        ret = script_runner.run(["hop", "list-topics", broker_url + ",".join(query_topics),
                                 "--no-auth"])

        assert ret.success
        assert ret.stderr == ""
        assert "No accessible topics" in ret.stdout

        mock_consumer.assert_called()
        for topic in query_topics:
            mock_consumer.return_value.list_topics.assert_any_call(topic=topic, timeout=-1.)

    # listing of specific topics, some of which exist and some of which do not
    with patch("confluent_kafka.Consumer", make_consumer_mock(topic_results)) as mock_consumer:
        ret = script_runner.run(["hop", "list-topics", broker_url + ",".join(query_topics),
                                 "--no-auth"])

        assert ret.success
        assert ret.stderr == ""
        assert "Accessible topics" in ret.stdout
        for topic in expected_topics:
            assert topic in ret.stdout
        for topic in unexpected_topics:
            assert topic not in ret.stdout

        mock_consumer.assert_called()
        for topic in query_topics:
            mock_consumer.return_value.list_topics.assert_any_call(topic=topic, timeout=-1.)

    # general listing with authentication
    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir), \
            patch("confluent_kafka.Consumer", make_consumer_mock(topic_results)) as mock_consumer:
        ret = script_runner.run(["hop", "list-topics", broker_url])

        assert ret.success
        assert ret.stderr == ""
        assert "Accessible topics" in ret.stdout
        for topic in expected_topics:
            assert topic in ret.stdout
        for topic in unexpected_topics:
            assert topic not in ret.stdout

        mock_consumer.assert_called()
        mock_consumer.return_value.list_topics.assert_called_with(timeout=-1.)

    # attempting to use multiple brokers should provoke an error
    ret = script_runner.run(["hop", "list-topics", "kafka://example.com,example.net"])
    assert not ret.success
    assert "Multiple broker addresses are not supported" in ret.stderr


def test_cli_configure_locate(script_runner, general_config, tmpdir):
    with temp_config(tmpdir, general_config) as conf_dir, temp_environ(XDG_CONFIG_HOME=conf_dir):
        ret1 = script_runner.run(["hop", "configure", "--help"])
        assert ret1.success
        assert ret1.stderr == ""

        ret = script_runner.run(["hop", "configure", "locate"])
        assert ret.success
        assert conf_dir in ret.stdout
        assert ret.stderr == ""


def test_cli_configure_show(script_runner, general_config, tmpdir):
    with temp_config(tmpdir, general_config) as conf_dir, temp_environ(XDG_CONFIG_HOME=conf_dir):
        ret = script_runner.run(["hop", "configure", "show"])
        assert ret.success
        assert "fetch_external: False" in ret.stdout
        assert "automatic_offload: False" in ret.stdout
        assert ret.stderr == ""


def test_cli_configure_set(script_runner, general_config, tmpdir):
    with temp_config(tmpdir, general_config) as conf_dir, temp_environ(XDG_CONFIG_HOME=conf_dir):
        ret = script_runner.run(["hop", "configure", "set", "fetch_external", "true"])
        assert ret.success
        config = configure.load_config()
        assert config.fetch_external

    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        ret = script_runner.run(["hop", "configure", "set", "not-a-setting", "18"])
        assert not ret.success
        assert "not-a-setting" in ret.stderr

    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        ret = script_runner.run(["hop", "configure", "set", "fetch_external", "warbler"])
        assert not ret.success
        assert "Invalid boolean value 'warbler'" in ret.stderr


def test_cli_auth(script_runner, auth_config, tmpdir):
    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        ret1 = script_runner.run(["hop", "auth", "--help"])
        assert ret1.success
        assert ret1.stderr == ""

        ret = script_runner.run(["hop", "auth", "locate"])
        assert ret.success
        assert config_dir in ret.stdout
        assert ret.stderr == ""


def test_list_credentials(script_runner, auth_config, tmpdir):
    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        ret = script_runner.run(["hop", "auth", "list"])
        assert ret.success
        assert "username" in ret.stdout
        assert ret.stderr == ""


def test_add_credential(script_runner, auth_config, tmpdir):
    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        csv_file = str(tmpdir) + "/new_cred.csv"
        with open(csv_file, "w") as f:
            f.write("username,password\nnew_user,new_pass")
        ret = script_runner.run(["hop", "auth", "add", csv_file])
        assert ret.success
        assert "Wrote configuration to" in ret.stderr


def test_add_credential_overwrite(script_runner, auth_config, tmpdir):
    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        csv_file = str(tmpdir) + "/new_cred.csv"
        with open(csv_file, "w") as f:
            f.write("username,password\nnew_user,new_pass")
        ret = script_runner.run(["hop", "auth", "add", csv_file])
        assert ret.success
        assert "Wrote configuration to" in ret.stderr

        with open(csv_file, "w") as f:
            f.write("username,password\nnew_user,other_pass")

        # try to overwrite the credential, without forcing
        ret = script_runner.run(["hop", "auth", "add", csv_file])
        assert ret.success
        assert "Credential already exists; overwrite with --force" in ret.stderr

        # try again, with force
        ret = script_runner.run(["hop", "auth", "add", "--force", csv_file])
        assert ret.success
        assert "Wrote configuration to" in ret.stderr


def test_delete_credential(script_runner, auth_config, tmpdir):
    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        ret = script_runner.run(["hop", "auth", "remove", "username"])
        assert ret.success
        assert "Wrote configuration to" in ret.stderr


def test_cli_version(script_runner, auth_config, tmpdir):
    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        ret = script_runner.run(["hop", "version", "--help"])
        assert ret.success
        assert ret.stderr == ""

        ret = script_runner.run(["hop", "version"])
        assert ret.success
        assert f"hop-client=={__version__}\n" in ret.stdout
        assert ret.stderr == ""


def test_error_verbosity(script_runner):
    simple = script_runner.run(["hop", "subscribe", "BAD-URL"])
    assert not simple.success
    assert simple.stdout == ""
    assert "Traceback (most recent call last)" not in simple.stderr
    assert simple.stderr.startswith("hop: ")

    detailed = script_runner.run(["hop", "--debug", "subscribe", "BAD-URL"])
    assert not detailed.success
    assert detailed.stdout == ""
    assert "Traceback (most recent call last)" in detailed.stderr


def test_config_advice(script_runner, auth_config, tmpdir):
    advice_tag = "No valid credential data found"
    # nonexistent config file
    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        ret = script_runner.run(["hop"])
        assert advice_tag in ret.stdout

    # wrong credential file permissions
    import stat
    with temp_auth(tmpdir, "", stat.S_IROTH) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir):
        ret = script_runner.run(["hop"])
        assert advice_tag in ret.stdout
        assert "unsafe permissions" in ret.stderr

    # syntactically invalid TOML in credential file
    garbage = "JVfwteouh '652b"
    with temp_auth(tmpdir, garbage) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        ret = script_runner.run(["hop"])
        assert advice_tag in ret.stdout
        assert "not configured correctly" in ret.stderr

    # syntactically valid TOML without an [auth] section
    toml_no_auth = """title = "TOML Example"
    [owner]
    name = "Tom Preston-Werner"
    dob = 1979-05-27T07:32:00-08:00
    """
    with temp_auth(tmpdir, toml_no_auth) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        ret = script_runner.run(["hop"])
        assert advice_tag in ret.stdout
        assert "configuration file has no auth section" in ret.stderr

    # syntactically valid TOML an incomplete [auth] section
    toml_bad_auth = """[auth]
    foo = "bar"
    """
    with temp_auth(tmpdir, toml_bad_auth) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        ret = script_runner.run(["hop"])
        assert advice_tag in ret.stdout
        assert "missing auth property" in ret.stderr



import os
from setuptools import find_packages, setup

# read in README
this_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_dir, 'README.md'), 'rb') as f:
    long_description = f.read().decode().strip()

# requirements
install_requires = [
    "adc-streaming >= 2.5.0",
    "dataclasses ; python_version < '3.7'",
    "fastavro >= 1.4.0",
    "pluggy >= 0.11",
    "toml >= 0.9.4",
    "xmltodict >= 0.9.0",
    "scramp",
    "bson",
]
extras_require = {
    'dev': [
        'autopep8',
        'flake8 >= 3.3.0',
        'pytest >= 6.2.5',
        'pytest-console-scripts',
        'pytest-cov',
        'pytest-runner',
        'twine',
    ],
    'docs': [
        'sphinx',
        'sphinx_rtd_theme',
        'sphinxcontrib-programoutput',
    ],
}

setup(
    name = 'hop-client',
    description = 'A pub-sub client library for Multi-messenger Astrophysics',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/scimma/hop-client',
    author = 'Patrick Godwin',
    author_email = 'patrick.godwin@psu.edu',
    license = 'BSD 3-Clause',

    packages = find_packages(),

    entry_points = {
        'console_scripts': [
            'hop = hop.__main__:main',
        ],
    },

    python_requires = '>=3.8',
    install_requires = install_requires,
    extras_require = extras_require,
    setup_requires = ['setuptools_scm'],
    zip_safe=False,
    use_scm_version = {
        'write_to': 'hop/_version.py'
    },

    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'License :: OSI Approved :: BSD License',
    ],

)



import subprocess
from hop import version


def test_version():
    process = subprocess.run(
        ['hop', 'version'], stdout=subprocess.PIPE, universal_newlines=True)

    versions = process.stdout
    packages = version.get_packages()
    assert all(package in versions for package in packages)
